/*
 *	Button functions
 */

#include <stdlib.h>
#include <stdio.h>
#include <vga.h>
#include <vgagl.h>
#include <vgamouse.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include "hr_defines.h"
#include "hr_colors.h"
#include "hr_fonts.h"
#include "hr_mouse.h"
#include "hr_bobs.h"
#include "hr_general.h"
#include "hr_bitmaps.h"
#include "hr_keyboard.h"
#include "hr_editor.h"
#include "hr_windows.h"
#include "hr_zones.h"
#include "hr_graphics.h"
#include "hr_strings.h"
#include "hr_screens.h"

/* When setting up sliders that are only horiz/vert and not both,
 * x/y variables can be ignored.*/
struct hr_slider{
	int flags;
	int x_items;
	int x_visible;
	int x_position;
	int x_minval;
	int *x_valint;
	int y_items;
	int y_visible;
	int y_position;
	int y_minval;
	int *y_valint;
};

/* Radio buttons.*/
struct hr_buttongroup{
	int flags;
	int buttons;
	struct hr_button **button;
};

/**/
struct hr_button{
	struct hr_button *next;
	int flags;
	struct hr_font  *font;
	struct hr_color *colors;
	unsigned char *text;
	unsigned char *text2;
	unsigned char *helptext;
	struct hr_bitmap *bitmap;
	struct hr_bitmap *bitmap2;
	struct hr_zone *zone;
	struct hr_window *parent_window;
	int hotkey;
	void *extras;
};


/* Adds button to a group*/
struct hr_buttongroup *hr_add_to_group(struct hr_buttongroup *group,
													struct hr_button *b)
{
	if(b!=NULL){
		if(group==NULL){
			group=malloc(sizeof(struct hr_buttongroup));
			group->flags=NONE;
			group->buttons=0;
			group->button=malloc(sizeof(struct hr_button *));
		}
		/* Make pointer array larger if necessary.*/
		if(++(group->buttons)>1)
		  group->button=realloc(group->button,
										sizeof(struct hr_button *)*(group->buttons));
	}
	return group;
}

/* Deletes a buttongroup */
struct hr_buttongroup *hr_delete_group(struct hr_buttongroup *group){
	int l;
	if(group!=NULL){
		/* NULLify the extras-pointers of every button belonging to
		 * this group.*/
		for(l=0;l<group->buttons;l++)
		  (((group->button)[l]))->extras=NULL;
		free(group->button);
		free(group);
	}
	return group;
}



/* Returns button under coordinates x,y or -1 if no button is there.*/
struct hr_button *hr_get_button(int x,int y,int flags)
{
	struct hr_button *b;
	struct hr_zone *z;
	z=hr.firstzone;
	/* Scan only zones that have ZONE_GADGET flag set.*/
	while(z!=NULL){
		if((z=hr_get_zone(z,x,y,ZONE_GADGET))){
			/* Check if any button->zone pointer points to this particular
			 * zone. If yes, return pointer to that button.*/
			b=hr.firstbutton;
			while(b!=NULL){
				if(b->zone == z){
					if((flags & BUTTON_IGNORE_ENABLED_FLAG)||
						(b->flags & BUTTON_ENABLED))
					  return b;
				}
				b=b->next;
			}
			/* Move z to next zone to prevent looping forever.. */
			z=z->next;
		}
		else
		  break;
	}
	return NULL;
}


/* Shades a rectangular area */
void hr_shade(int x1,int y1,int x2,int y2,struct hr_color *colors,
				  int shadecolor)
{
	int x,y;
	for(y=y1;y<=y2;y++)  {
		for(x=x1;x<=x2;x++)  {
			if((x%2)==(y%2)){
				gl_setpixel(x,y,hr_get_color_index(colors,shadecolor));
			}
		}
	}
}


/* Clears button gfx on screen.*/
void hr_clear_button(struct hr_button *p)
{
	if(p!=NULL)
	  gl_fillbox((p->zone->x1)+1,(p->zone->y1)+1,
					 abs((p->zone->x2)-(p->zone->x1))-1,
					 abs((p->zone->y2)-(p->zone->y1))-1,
					 (p->flags & BUTTON_PRESSED)?
					 (hr_get_color_index(p->colors,COLOR_PRESSED)):
					 (hr_get_color_index(p->colors,COLOR_RAISED)));
}


/* If pressed==1, the textbox is darker gray,
 * otherwise brighter gray (compared to color 0).
 * Text that doesn't fit in the button is cropped.
 * NOTE! Font height may exceed height of button - no checking is made! */
void hr_textbox(int x1,int y1,int x2,int y2,
					 struct hr_font *font,
					 struct hr_color *colors,
					 unsigned char *text,int offset,int len,int flags)
{
	int w,fc,bc;
	int cx1,cy1,cx2,cy2;
	/* Clear background and draw frame */
	if(flags & BUTTON_HELPTEXT){
		gl_fillbox(x1+1,y1+1,abs(x2-x1)-1,abs(y2-y1)-1,
					  hr_get_color_index(colors,COLOR_DARK));
		hr_box(x1,y1,x2,y2,hr_get_color_index(colors,COLOR_HILITE),NONE);
		fc=hr_get_color_index(colors,COLOR_HILITE);
	}
	else{
		gl_fillbox(x1+1,y1+1,abs(x2-x1)-1,abs(y2-y1)-1,
					  ((flags & BUTTON_PRESSED)?
						(hr_get_color_index(colors,COLOR_PRESSED)):
						(hr_get_color_index(colors,COLOR_RAISED))));
		hr_bevelbox(x1,y1,x2,y2,colors,flags);
		fc=hr_get_color_index(colors,COLOR_FOREGROUND);
	}
	/* Print the text. */
	if(text!=NULL)  {
		bc=hr_get_color_index(colors,COLOR_BACKGROUND);
		/**/
		cx1=__clipx1;
		cy1=__clipy1;
		cx2=__clipx2;
		cy2=__clipy2;
		gl_setclippingwindow(x1+1,y1+1,x2-1,y2-1);
		/**/
		w=abs(x2-x1)+1;
		hr_write((x1+x2-hr_string_xpixels(text,font,offset,len,NONE))/2,
					((y1+y2-(font->height))/2)+1,
					fc,bc,text,font,offset,len,NONE);
		/**/
		gl_setclippingwindow(cx1,cy1,cx2,cy2);
	}
	/* Shade the textbox a la Amiga if it is disabled */
	if(!(flags & BUTTON_ENABLED))
	  hr_shade(x1,y1,x2,y2,colors,COLOR_DARK);
}


/* Prints a text with accompanied textbox with a string in it.
 *                            +-----+
 * --->                  text |value|
 *                            +-----+ */
void hr_value_gadget(int x,int y,int w,int h,int offset,
							struct hr_font *font,
							struct hr_color *colors,
							unsigned char *text,
							unsigned char *valuetext,int value,int flags)
{
	unsigned char tmp[HR_BUFFER_SIZE];
	if(text!=NULL){
		strncpy(tmp,text,HR_BUFFER_SIZE-1);
		tmp[offset/8]=0;
		hr_write(x,y+(h/2)-4,
					hr_get_color_index(colors,COLOR_FOREGROUND),
					hr_get_color_index(colors,COLOR_BACKGROUND),tmp,font,
					STRING_START,STRING_END,NONE);
	}
	if(valuetext==NULL) {
		sprintf(tmp,"%d",value);
		valuetext=tmp;
	}
	hr_textbox(x+offset,y,x+w-1,y+h-1,
				  font,colors,
				  valuetext,STRING_START,STRING_END,flags);
}


/* Unneccessary func, but makes source look nicer.*/
struct hr_slider *hr_create_slider(int flags,
											  int x_items,int x_visible,int x_position,
											  int x_minval,int *x_valint,
											  int y_items,int y_visible,int y_position,
											  int y_minval,int *y_valint)
{
	struct hr_slider *s=NULL;
	s=malloc(sizeof(struct hr_slider));
	s->flags=flags;
	s->x_items=MAX(1,x_items);
	s->x_visible=MIN(x_visible,x_items);/* Just to be sure :-) */
	s->x_position=x_position;
	s->x_minval=x_minval;
	s->x_valint=x_valint;
	s->y_items=MAX(1,y_items);
	s->y_visible=MIN(y_visible,y_items);/* Just to be sure :-) */
	s->y_position=y_position;
	s->y_minval=y_minval;
	s->y_valint=y_valint;
	return s;
}


/* This adds a 3D slider bar to a button that has its
 * BUTTON_HSLIDER or/and BUTTON_VSLIDER flags set.
 * Should be called from within update_button() only.*/
void hr_slider_bar(struct hr_button *b)
{
	struct hr_slider *s=(struct hr_slider *)(b->extras);
	int x1,y1,x2,y2,l;
	if(s!=NULL){
		/* Set shitloads of checks! FIX! FIX! FIX! FIX!
		 * w and/or h may be negative!
		 * default positions of topleft and bottomright corners of the
		 * slider bar */
		x1=(b->zone->x1)+1;
		y1=(b->zone->y1)+1;
		x2=(b->zone->x2)-1;
	   y2=(b->zone->y2)-1;
		/* If the slider button is set to watch either xvalint or yvalint,
		 * set new x/y_positions. This allows very easy updating
		 * of interactive sliders.*/
		if(s->flags & SLIDER_GET_XVALINT){
			if(s->x_valint!=NULL){
				if(s->flags & SLIDER_XNEG)
				  s->x_position=(s->x_minval)-*(s->x_valint);
				else
				  s->x_position=*(s->x_valint)-(s->x_minval);
			}
		}
		if(s->flags & SLIDER_GET_YVALINT){
			if(s->y_valint!=NULL){
				if(s->flags & SLIDER_YNEG)
				  s->y_position=(s->y_minval)-*(s->y_valint);
				else
				  s->y_position=*(s->y_valint)-(s->y_minval);
			}
		}
		if(b->flags & BUTTON_HSLIDER){
			if((s->x_items)>1){
				/* Calc width of bar */
				l=((s->x_visible)*((b->zone->x2)-(b->zone->x1)-1))/(s->x_items);
				/* Calc x1/x2.*/
				x1=(b->zone->x1)+1+(((s->x_position)*
											((b->zone->x2)-
											 (b->zone->x1)-1-l))/((s->x_items)-1));
				x2=x1+l;
				if(x1<=(b->zone->x1))
				  x1=(b->zone->x1)+1;
				if(x2>=(b->zone->x2))
				  x2=(b->zone->x2)-1;
			}
		}
		if(b->flags & BUTTON_VSLIDER){
			if((s->y_items)>1){
				/* Calc height of bar */
				l=((s->y_visible)*((b->zone->y2)-(b->zone->y1)-1))/(s->y_items);
				/* Calc y1/y2*/
				y1=(b->zone->y1)+1+(((s->y_position)*
									((b->zone->y2)-
									 (b->zone->y1)-1-l))/((s->y_items)-1));
				y2=y1+l;
				if(y1<=(b->zone->y1))
				  y1=(b->zone->y1)+1;
				if(y2>=(b->zone->y2))
				  y2=(b->zone->y2)-1;
			}
		}
		/* Draw it.*/
		gl_fillbox(x1,y1,(x2-x1+1),(y2-y1+1),
					  hr_get_color_index(b->colors,COLOR_BACKGROUND));
		hr_bevelbox(x1,y1,x2-1,y2-1,b->colors,NONE);
	}
}


/* Updates button condition and possibly changes texts.
 * Bitmaps can be changed too using this func...
 * So animated buttons are possible :-) */
void hr_update_button(struct hr_button *p,int mode,void *data)
{
	int w,clear_background=1;
	struct hr_bitmap *tb;
	unsigned char *tt;
	int cx1,cy1,cx2,cy2,cf;
	/**/
	if(p!=NULL) {
		if(data!=NULL){
			/* If a new string is given, a copy of the text is made.
			 * Otherwise just pointers are set.*/
			switch(mode){
#if 0
			 case UPDATE_GFXONLY:
				break;
#endif
			 case UPDATE_FLAGS:
				p->flags=(*(int *)data);
				break;
			 case UPDATE_X1:
				hr_clear_button(p);
				p->zone->x1=(*(int *)data);
				clear_background=0;
				break;
			 case UPDATE_Y1:
				hr_clear_button(p);
				p->zone->y1=(*(int *)data);
				clear_background=0;
				break;
			 case UPDATE_X2:
				hr_clear_button(p);
				p->zone->x2=(*(int *)data);
				clear_background=0;
				break;
			 case UPDATE_Y2:
				hr_clear_button(p);
				p->zone->y2=(*(int *)data);
				clear_background=0;
				break;
			 case UPDATE_TEXT:
				if((unsigned char *)data!=(p->text)) {
					free(p->text);
					p->text=malloc(strlen((unsigned char *)data)+1);
					strcpy(p->text,(unsigned char *)data);	
				}
				break;
			 case UPDATE_TEXT2:
				if((unsigned char *)data!=(p->text2)) {
					free(p->text2);
					p->text2=malloc(strlen((unsigned char *)data)+1);
					strcpy(p->text2,(unsigned char *)data);
				}
				break;
			 case UPDATE_HELPTEXT:
				if((unsigned char *)data!=(p->helptext)) {
					free(p->helptext);
					p->helptext=malloc(strlen((unsigned char *)data)+1);
					strcpy(p->helptext,(unsigned char *)data);
				}
				break;
			 case UPDATE_BITMAP:
				/* Note that "old" bitmap is not freed!*/
				p->bitmap=(struct hr_bitmap *)data;
				break;
			 case UPDATE_BITMAP2:
				/* Note that "old" bitmap is not freed!*/
				p->bitmap2=(struct hr_bitmap *)data;
				break;
			 case UPDATE_FONT:
				/* Font modifying. */
				p->font=(struct hr_font *)data;
				break;
			 case UPDATE_COLORS:
				/* Color modifying */
				p->colors=(struct hr_color *)data;
				break;
			 case UPDATE_HOTKEY:
				/* Change hotkey for button */
				p->hotkey=(int)(*((int *)data));
				break;
			 case UPDATE_EXTRAS:
				/* Set extras-pointer.*/
				p->extras=data;
				break;
			}
		}
		/* Draw the button if it is not invisible.
		 * The check is made here because it makes no sense to create
		 * invisible textboxes :-) */
		if(!(p->flags & BUTTON_INVISIBLE)){
			/* Clear the background.*/
			if(clear_background)
			  hr_clear_button(p);
			/* Draw the bitmap.*/
			tb=(p->flags & BUTTON_PRESSED)?(p->bitmap2):(p->bitmap);
			cf=__clip;
			cx1=__clipx1;
			cy1=__clipy1;
			cx2=__clipx2;
			cy2=__clipy2;
			if(!cf)
			  gl_enableclipping();
			gl_setclippingwindow(p->zone->x1,p->zone->y1,
										p->zone->x2,p->zone->y2);
			if(tb!=NULL)
			  gl_putboxmask(((p->zone->x1)+(p->zone->x2))/2-(tb->width)/2,
								 ((p->zone->y1)+(p->zone->y2))/2-(tb->height)/2,
								 tb->width,tb->height,tb->data);
			/* Hilite? */
			if((hr.default_button==p) &&
				(p->flags & BUTTON_ENABLED))
			  gl_hline((p->zone->x1)+1,(p->zone->y2)-1,
						  (p->zone->x2)-1,hr_get_color_index(p->colors,COLOR_HILITE));
			/* Print the text. If text2 does not exist, text is always used.*/
			if((p->text2)!=NULL)
			  tt=(p->flags & BUTTON_PRESSED)?(p->text2):(p->text);
			else
			  tt=p->text;
			if(tt!=NULL){
				w=abs((p->zone->x2)-(p->zone->x1))+1;
				/* Write the string.*/
				hr_write(((p->zone->x1)+(p->zone->x2)-
							 hr_string_xpixels(tt,p->font,
													 STRING_START,STRING_END,NONE))/2,
							((p->zone->y1)+(p->zone->y2)-(p->font->height))/2+1,
							hr_get_color_index(p->colors,COLOR_FOREGROUND),
							hr_get_color_index(p->colors,COLOR_BACKGROUND),
							tt,p->font,STRING_START,STRING_END,NONE);
			}
			/* If the button is a slider, draw the bar.*/
			if((p->flags & BUTTON_HSLIDER)||(p->flags & BUTTON_VSLIDER))
			  hr_slider_bar(p);
			/* Draw the 3D bevelbox */
			hr_bevelbox((p->zone->x1),(p->zone->y1),
							(p->zone->x2),(p->zone->y2),p->colors,p->flags);
			/* Draw inner box if the button is a radiobutton.
			 * This bevelbox is always pressed and enabled,
			 * and sometimes filled.*/
			if(p->flags & BUTTON_RADIO){
				w=(p->zone->y2)-(p->zone->y1)-3;
				hr_bevelbox((p->zone->x1+2),(p->zone->y1+2),
								(p->zone->x1)+1+w,(p->zone->y1)+1+w,p->colors,
								BUTTON_PRESSED|BUTTON_ENABLED);
				if(p->flags & BUTTON_PRESSED)
				  gl_fillbox((p->zone->x1)+3,(p->zone->y1)+3,
								 w-2,w-2,
								 hr_get_color_index(p->colors,COLOR_HILITE));
			}
			/* Shade the textbox a la Amiga if it is disabled */
			if(!(p->flags & BUTTON_ENABLED))
			  hr_shade(p->zone->x1,p->zone->y1,
						  p->zone->x2,p->zone->y2,p->colors,COLOR_DARK);
			/* Restore clipping.*/
			gl_setclippingwindow(cx1,cy1,cx2,cy2);
			if(!cf)
			  gl_disableclipping();
		}
	}
}


/* Creates a button and draws it to screen.
 * Returns pointer to the new button.
 * Note! If you create 1*1 pixel buttons, some other func may crash :-) */
struct hr_button *hr_create_button(int x1,int y1,int x2,int y2,
											  struct hr_font *font,
											  struct hr_color *colors,
											  unsigned char *text,
											  unsigned char *text2,
											  unsigned char hotkey,
											  int flags)
{
	struct hr_button *p=hr.firstbutton;
	if(p==NULL)
	  hr.firstbutton=p=malloc(sizeof(struct hr_button));
	else {			  
		while(p->next!=NULL)	
		  p=p->next;
		p->next=malloc(sizeof(struct hr_button));
		p=p->next;
	}
	/* The rest...*/
	p->next=NULL;
	p->bitmap=NULL;
	p->bitmap2=NULL;
	p->extras=NULL;
	p->hotkey=0;
	p->flags=flags;
	p->font=font;	
	p->colors=colors;
	p->zone=hr_create_zone(x1,y1,x2,y2,ZONE_GADGET);
	/* Make copy of texts accompanied with a button.
	 * This space is automatically freed when either the text is changed
	 * or the button removed.*/
	if(text!=NULL) {
		p->text=malloc(strlen(text)+1);;
		strcpy(p->text,text);
	}
	else
	  p->text=NULL;
	/* text2 */
	if(text2!=NULL) {
		p->text2=malloc(strlen(text2)+1);
		strcpy(p->text2,text2);
	}
	else
	  p->text2=NULL;
	p->helptext=NULL;
	/*parent window*/
	p->parent_window=hr.active_screen->active_window;
	/* Hotkey */
	p->hotkey=hotkey;
	/* Drawing is done in update_button */
	hr_update_button(p,UPDATE_GFXONLY,NULL);
	return p;
}


/* Removes a button from the button list (but NOT from screen!).
 * Button bitmaps are left untouched.*/
struct hr_button *hr_delete_button(struct hr_button *p)
{
	struct hr_button *p2;
	if(p!=NULL) {		
		/* Set pointers so that list is not screwed.*/
		if(p==hr.firstbutton)
		  hr.firstbutton=p->next;
		else {		
			p2=hr.firstbutton;
			while(p2->next != p)
			  p2=p2->next;
			p2->next=p->next;						/* Jump "over" p  */
		}
		/* Button string(s) are freed.
		 * Bitmaps, font and colors are left untouched.
		 * free(NULL) is ok. */
		hr_delete_zone(p->zone);
		free(p->text);
		free(p->text2);
		free(p->helptext);
		free(p->extras);
		free(p);
		/* Final check to avoid nasty things */
		if(hr.default_button==p)
		  hr.default_button=NULL;
		return p;
	}
	return NULL;
}
